@@ -399,12 +399,67 @@ grub_video_fb_set_palette (unsigned int start, unsigned int count,
   return GRUB_ERR_NONE;
 }
 
+static grub_err_t
+grub_video_fb_set_area (void)
+{
+  unsigned int viewport_x1 = framebuffer.render_target->viewport.x;
+  unsigned int viewport_y1 = framebuffer.render_target->viewport.y;
+  unsigned int viewport_width = framebuffer.render_target->viewport.width;
+  unsigned int viewport_height = framebuffer.render_target->viewport.height;
+  unsigned int viewport_x2 = viewport_x1 + viewport_width;
+  unsigned int viewport_y2 = viewport_y1 + viewport_height;
+
+  unsigned int region_x1 = framebuffer.render_target->region.x;
+  unsigned int region_y1 = framebuffer.render_target->region.y;
+  unsigned int region_width = framebuffer.render_target->region.width;
+  unsigned int region_height = framebuffer.render_target->region.height;
+  unsigned int region_x2 = region_x1 + region_width;
+  unsigned int region_y2 = region_y1 + region_height;
+
+  unsigned int max_x1 = grub_max (viewport_x1, region_x1);
+  unsigned int min_x2 = grub_min (viewport_x2, region_x2);
+  unsigned int max_y1 = grub_max (viewport_y1, region_y1);
+  unsigned int min_y2 = grub_min (viewport_y2, region_y2);
+
+  /* Viewport and region do not intersect. */
+  if (viewport_width == 0 || viewport_height == 0 || region_width == 0
+      || region_height == 0 || max_x1 >= min_x2 || max_y1 >= min_y2)
+    {
+      framebuffer.render_target->area.x = 0;
+      framebuffer.render_target->area.y = 0;
+      framebuffer.render_target->area.width = 0;
+      framebuffer.render_target->area.height = 0;
+      framebuffer.render_target->area_offset_x = 0;
+      framebuffer.render_target->area_offset_y = 0;
+      return GRUB_ERR_NONE;
+    }
+
+  /* There is non-zero intersection. */
+  framebuffer.render_target->area.x = max_x1;
+  framebuffer.render_target->area.y = max_y1;
+  framebuffer.render_target->area.width = min_x2 - max_x1;
+  framebuffer.render_target->area.height = min_y2 - max_y1;
+
+  if (region_x1 > viewport_x1)
+    framebuffer.render_target->area_offset_x = (int)region_x1
+                                               - (int)viewport_x1;
+  else
+    framebuffer.render_target->area_offset_x = 0;
+  if (region_y1 > viewport_y1)
+    framebuffer.render_target->area_offset_y = (int)region_y1
+                                               - (int)viewport_y1;
+  else
+    framebuffer.render_target->area_offset_y = 0;
+
+  return GRUB_ERR_NONE;
+}
+
 grub_err_t
 grub_video_fb_set_viewport (unsigned int x, unsigned int y,
 			    unsigned int width, unsigned int height)
 {
-  /* Make sure viewport is withing screen dimensions.  If viewport was set
-     to be out of the region, mark its size as zero.  */
+  /* Make sure viewport is within screen dimensions.  If viewport was set
+     to be out of the screen, mark its size as zero.  */
   if (x > framebuffer.render_target->mode_info.width)
     {
       x = 0;
@@ -428,6 +483,10 @@ grub_video_fb_set_viewport (unsigned int x, unsigned int y,
   framebuffer.render_target->viewport.width = width;
   framebuffer.render_target->viewport.height = height;
 
+  /* Count drawing area only if needed. */
+  if (framebuffer.render_target->area_enabled)
+    grub_video_fb_set_area ();
+
   return GRUB_ERR_NONE;
 }
 
@@ -443,6 +502,78 @@ grub_video_fb_get_viewport (unsigned int *x, unsigned int *y,
   return GRUB_ERR_NONE;
 }
 
+grub_err_t
+grub_video_fb_set_region (unsigned int x, unsigned int y,
+                          unsigned int width, unsigned int height)
+{
+  /* Make sure region is within screen dimensions.  If region was set
+     to be out of the screen, mark its size as zero.  */
+  if (x > framebuffer.render_target->mode_info.width)
+    {
+      x = 0;
+      width = 0;
+    }
+
+  if (y > framebuffer.render_target->mode_info.height)
+    {
+      y = 0;
+      height = 0;
+    }
+
+  if (x + width > framebuffer.render_target->mode_info.width)
+    width = framebuffer.render_target->mode_info.width - x;
+
+  if (y + height > framebuffer.render_target->mode_info.height)
+    height = framebuffer.render_target->mode_info.height - y;
+
+  framebuffer.render_target->region.x = x;
+  framebuffer.render_target->region.y = y;
+  framebuffer.render_target->region.width = width;
+  framebuffer.render_target->region.height = height;
+
+  /* If we have called set_region then area is needed.  */
+  grub_video_fb_set_area ();
+
+  return GRUB_ERR_NONE;
+}
+
+grub_err_t
+grub_video_fb_get_region (unsigned int *x, unsigned int *y,
+                          unsigned int *width, unsigned int *height)
+{
+  if (x) *x = framebuffer.render_target->region.x;
+  if (y) *y = framebuffer.render_target->region.y;
+  if (width) *width = framebuffer.render_target->region.width;
+  if (height) *height = framebuffer.render_target->region.height;
+
+  return GRUB_ERR_NONE;
+}
+
+grub_err_t
+grub_video_fb_set_area_status (grub_video_area_status_t area_status)
+{
+  if (area_status == GRUB_VIDEO_AREA_ENABLED)
+    framebuffer.render_target->area_enabled = 1;
+  else
+    framebuffer.render_target->area_enabled = 0;
+
+  return GRUB_ERR_NONE;
+}
+
+grub_err_t
+grub_video_fb_get_area_status (grub_video_area_status_t *area_status)
+{
+  if (!area_status)
+    return GRUB_ERR_NONE;
+
+  if (framebuffer.render_target->area_enabled)
+    *area_status = GRUB_VIDEO_AREA_ENABLED;
+  else
+    *area_status = GRUB_VIDEO_AREA_DISABLED;
+
+  return GRUB_ERR_NONE;
+}
+
 /* Maps color name to target optimized color format.  */
 grub_video_color_t
 grub_video_fb_map_color (grub_uint32_t color_name)
@@ -612,7 +743,7 @@ grub_video_fb_unmap_color_int (struct grub_video_fbblit_info * source,
   if ((mode_info->mode_type
        & GRUB_VIDEO_MODE_TYPE_INDEX_COLOR) != 0)
     {
-      if ((framebuffer.render_target->mode_info.mode_type
+      if ((mode_info->mode_type
 	   & GRUB_VIDEO_MODE_TYPE_ALPHA) != 0
 	  && color == 0xf0)
         {
@@ -714,14 +845,36 @@ grub_video_fb_fill_rect (grub_video_color_t color, int x, int y,
 			 unsigned int width, unsigned int height)
 {
   struct grub_video_fbblit_info target;
+  unsigned int area_x;
+  unsigned int area_y;
+  unsigned int area_width;
+  unsigned int area_height;
+  if (framebuffer.render_target->area_enabled)
+    {
+      area_x = framebuffer.render_target->area.x;
+      area_y = framebuffer.render_target->area.y;
+      area_width = framebuffer.render_target->area.width;
+      area_height = framebuffer.render_target->area.height;
+      x -= framebuffer.render_target->area_offset_x;
+      y -= framebuffer.render_target->area_offset_y;
+    }
+  else
+    {
+      area_x = framebuffer.render_target->viewport.x;
+      area_y = framebuffer.render_target->viewport.y;
+      area_width = framebuffer.render_target->viewport.width;
+      area_height = framebuffer.render_target->viewport.height;
+    }
 
   /* Make sure there is something to do.  */
-  if ((x >= (int)framebuffer.render_target->viewport.width) || (x + (int)width < 0))
+  if ((area_width == 0) || (area_height == 0))
     return GRUB_ERR_NONE;
-  if ((y >= (int)framebuffer.render_target->viewport.height) || (y + (int)height < 0))
+  if ((x >= (int)area_width) || (x + (int)width < 0))
+    return GRUB_ERR_NONE;
+  if ((y >= (int)area_height) || (y + (int)height < 0))
     return GRUB_ERR_NONE;
 
-  /* Do not allow drawing out of viewport.  */
+  /* Do not allow drawing out of area.  */
   if (x < 0)
     {
       width += x;
@@ -733,14 +886,14 @@ grub_video_fb_fill_rect (grub_video_color_t color, int x, int y,
       y = 0;
     }
 
-  if ((x + width) > framebuffer.render_target->viewport.width)
-    width = framebuffer.render_target->viewport.width - x;
-  if ((y + height) > framebuffer.render_target->viewport.height)
-    height = framebuffer.render_target->viewport.height - y;
+  if ((x + width) > area_width)
+    width = area_width - x;
+  if ((y + height) > area_height)
+    height = area_height - y;
 
-  /* Add viewport offset.  */
-  x += framebuffer.render_target->viewport.x;
-  y += framebuffer.render_target->viewport.y;
+  /* Add area offset.  */
+  x += area_x;
+  y += area_y;
 
   dirty (y, height);
 
@@ -753,30 +906,49 @@ grub_video_fb_fill_rect (grub_video_color_t color, int x, int y,
   return GRUB_ERR_NONE;
 }
 
-grub_err_t
-grub_video_fb_blit_bitmap (struct grub_video_bitmap *bitmap,
-			   enum grub_video_blit_operators oper, int x, int y,
-			   int offset_x, int offset_y,
-			   unsigned int width, unsigned int height)
+static inline grub_err_t __attribute__ ((always_inline))
+grub_video_fb_blit_source (struct grub_video_fbblit_info *source,
+                           enum grub_video_blit_operators oper, int x, int y,
+                           int offset_x, int offset_y,
+                           unsigned int width, unsigned int height)
 {
-  struct grub_video_fbblit_info source;
   struct grub_video_fbblit_info target;
+  unsigned int area_x;
+  unsigned int area_y;
+  unsigned int area_width;
+  unsigned int area_height;
+  if (framebuffer.render_target->area_enabled)
+    {
+      area_x = framebuffer.render_target->area.x;
+      area_y = framebuffer.render_target->area.y;
+      area_width = framebuffer.render_target->area.width;
+      area_height = framebuffer.render_target->area.height;
+      x -= framebuffer.render_target->area_offset_x;
+      y -= framebuffer.render_target->area_offset_y;
+    }
+  else
+    {
+      area_x = framebuffer.render_target->viewport.x;
+      area_y = framebuffer.render_target->viewport.y;
+      area_width = framebuffer.render_target->viewport.width;
+      area_height = framebuffer.render_target->viewport.height;
+    }
 
   /* Make sure there is something to do.  */
-  if ((width == 0) || (height == 0))
+  if ((area_width == 0) || (area_height == 0) || (width == 0) || (height == 0))
     return GRUB_ERR_NONE;
-  if ((x >= (int)framebuffer.render_target->viewport.width) || (x + (int)width < 0))
+  if ((x >= (int)area_width) || (x + (int)width < 0))
     return GRUB_ERR_NONE;
-  if ((y >= (int)framebuffer.render_target->viewport.height) || (y + (int)height < 0))
+  if ((y >= (int)area_height) || (y + (int)height < 0))
     return GRUB_ERR_NONE;
-  if ((x + (int)bitmap->mode_info.width) < 0)
+  if ((x + (int)source->mode_info->width) < 0)
     return GRUB_ERR_NONE;
-  if ((y + (int)bitmap->mode_info.height) < 0)
+  if ((y + (int)source->mode_info->height) < 0)
     return GRUB_ERR_NONE;
-  if ((offset_x >= (int)bitmap->mode_info.width)
+  if ((offset_x >= (int)source->mode_info->width)
       || (offset_x + (int)width < 0))
     return GRUB_ERR_NONE;
-  if ((offset_y >= (int)bitmap->mode_info.height)
+  if ((offset_y >= (int)source->mode_info->height)
       || (offset_y + (int)height < 0))
     return GRUB_ERR_NONE;
 
@@ -809,132 +981,66 @@ grub_video_fb_blit_bitmap (struct grub_video_bitmap *bitmap,
       y = 0;
     }
 
-  /* Do not allow drawing out of viewport.  */
-  if ((x + width) > framebuffer.render_target->viewport.width)
-    width = framebuffer.render_target->viewport.width - x;
-  if ((y + height) > framebuffer.render_target->viewport.height)
-    height = framebuffer.render_target->viewport.height - y;
+  /* Do not allow drawing out of area.  */
+  if ((x + width) > area_width)
+    width = area_width - x;
+  if ((y + height) > area_height)
+    height = area_height - y;
 
-  if ((offset_x + width) > bitmap->mode_info.width)
-    width = bitmap->mode_info.width - offset_x;
-  if ((offset_y + height) > bitmap->mode_info.height)
-    height = bitmap->mode_info.height - offset_y;
+  if ((offset_x + width) > source->mode_info->width)
+    width = source->mode_info->width - offset_x;
+  if ((offset_y + height) > source->mode_info->height)
+    height = source->mode_info->height - offset_y;
 
   /* Limit drawing to source render target dimensions.  */
-  if (width > bitmap->mode_info.width)
-    width = bitmap->mode_info.width;
+  if (width > source->mode_info->width)
+    width = source->mode_info->width;
 
-  if (height > bitmap->mode_info.height)
-    height = bitmap->mode_info.height;
+  if (height > source->mode_info->height)
+    height = source->mode_info->height;
 
   /* Add viewport offset.  */
-  x += framebuffer.render_target->viewport.x;
-  y += framebuffer.render_target->viewport.y;
+  x += area_x;
+  y += area_y;
 
   /* Use fbblit_info to encapsulate rendering.  */
-  source.mode_info = &bitmap->mode_info;
-  source.data = bitmap->data;
   target.mode_info = &framebuffer.render_target->mode_info;
   target.data = framebuffer.render_target->data;
 
   /* Do actual blitting.  */
   dirty (y, height);
-  grub_video_fb_dispatch_blit (&target, &source, oper, x, y, width, height,
-				       offset_x, offset_y);
+  grub_video_fb_dispatch_blit (&target, source, oper, x, y, width, height,
+                               offset_x, offset_y);
 
   return GRUB_ERR_NONE;
 }
 
 grub_err_t
-grub_video_fb_blit_render_target (struct grub_video_fbrender_target *source,
-                                   enum grub_video_blit_operators oper,
-                                   int x, int y, int offset_x, int offset_y,
-                                   unsigned int width, unsigned int height)
+grub_video_fb_blit_bitmap (struct grub_video_bitmap *bitmap,
+                           enum grub_video_blit_operators oper, int x, int y,
+                           int offset_x, int offset_y,
+                           unsigned int width, unsigned int height)
 {
   struct grub_video_fbblit_info source_info;
-  struct grub_video_fbblit_info target_info;
+  source_info.mode_info = &bitmap->mode_info;
+  source_info.data = bitmap->data;
 
-  /* Make sure there is something to do.  */
-  if ((width == 0) || (height == 0))
-    return GRUB_ERR_NONE;
-  if ((x >= (int)framebuffer.render_target->viewport.width) || (x + (int)width < 0))
-    return GRUB_ERR_NONE;
-  if ((y >= (int)framebuffer.render_target->viewport.height) || (y + (int)height < 0))
-    return GRUB_ERR_NONE;
-  if ((x + (int)source->mode_info.width) < 0)
-    return GRUB_ERR_NONE;
-  if ((y + (int)source->mode_info.height) < 0)
-    return GRUB_ERR_NONE;
-  if ((offset_x >= (int)source->mode_info.width)
-      || (offset_x + (int)width < 0))
-    return GRUB_ERR_NONE;
-  if ((offset_y >= (int)source->mode_info.height)
-      || (offset_y + (int)height < 0))
-    return GRUB_ERR_NONE;
+  return grub_video_fb_blit_source (&source_info, oper, x, y,
+                                    offset_x, offset_y, width, height);
+}
 
-  /* If we have negative coordinates, optimize drawing to minimum.  */
-  if (offset_x < 0)
-    {
-      width += offset_x;
-      x -= offset_x;
-      offset_x = 0;
-    }
-
-  if (offset_y < 0)
-    {
-      height += offset_y;
-      y -= offset_y;
-      offset_y = 0;
-    }
-
-  if (x < 0)
-    {
-      width += x;
-      offset_x -= x;
-      x = 0;
-    }
-
-  if (y < 0)
-    {
-      height += y;
-      offset_y -= y;
-      y = 0;
-    }
-
-  /* Do not allow drawing out of viewport.  */
-  if ((x + width) > framebuffer.render_target->viewport.width)
-    width = framebuffer.render_target->viewport.width - x;
-  if ((y + height) > framebuffer.render_target->viewport.height)
-    height = framebuffer.render_target->viewport.height - y;
-
-  if ((offset_x + width) > source->mode_info.width)
-    width = source->mode_info.width - offset_x;
-  if ((offset_y + height) > source->mode_info.height)
-    height = source->mode_info.height - offset_y;
-
-  /* Limit drawing to source render target dimensions.  */
-  if (width > source->mode_info.width)
-    width = source->mode_info.width;
-
-  if (height > source->mode_info.height)
-    height = source->mode_info.height;
-
-  /* Add viewport offset.  */
-  x += framebuffer.render_target->viewport.x;
-  y += framebuffer.render_target->viewport.y;
-
-  /* Use fbblit_info to encapsulate rendering.  */
+grub_err_t
+grub_video_fb_blit_render_target (struct grub_video_fbrender_target *source,
+                                  enum grub_video_blit_operators oper,
+                                  int x, int y, int offset_x, int offset_y,
+                                  unsigned int width, unsigned int height)
+{
+  struct grub_video_fbblit_info source_info;
   source_info.mode_info = &source->mode_info;
   source_info.data = source->data;
-  target_info.mode_info = &framebuffer.render_target->mode_info;
-  target_info.data = framebuffer.render_target->data;
 
-  /* Do actual blitting.  */
-  dirty (y, height);
-  grub_video_fb_dispatch_blit (&target_info, &source_info, oper, x, y, width, height,
-			       offset_x, offset_y);
-
-  return GRUB_ERR_NONE;
+  return grub_video_fb_blit_source (&source_info, oper, x, y,
+                                    offset_x, offset_y, width, height);
 }
 
 grub_err_t
@@ -1125,12 +1231,25 @@ grub_video_fb_create_render_target (struct grub_video_fbrender_target **result,
   /* Mark render target as allocated.  */
   target->is_allocated = 1;
 
-  /* Maximize viewport.  */
+  /* Maximize viewport, region and area.  */
   target->viewport.x = 0;
   target->viewport.y = 0;
   target->viewport.width = width;
   target->viewport.height = height;
 
+  target->region.x = 0;
+  target->region.y = 0;
+  target->region.width = width;
+  target->region.height = height;
+
+  target->area_enabled = 0;
+  target->area.x = 0;
+  target->area.y = 0;
+  target->area.width = width;
+  target->area.height = height;
+  target->area_offset_x = 0;
+  target->area_offset_y = 0;
+
   /* Setup render target format.  */
   target->mode_info.width = width;
   target->mode_info.height = height;
@@ -1175,7 +1294,11 @@ grub_video_fb_create_render_target (struct grub_video_fbrender_target **result,
     }
 
   /* Clear render target with black and maximum transparency.  */
-  grub_memset (target->data, 0, size);
+  if (mode_type == (GRUB_VIDEO_MODE_TYPE_INDEX_COLOR
+		    | GRUB_VIDEO_MODE_TYPE_ALPHA))
+    grub_memset (target->data, 0xf0, size);
+  else
+    grub_memset (target->data, 0, size);
 
   /* TODO: Add render target to render target list.  */
 
@@ -1213,12 +1336,25 @@ grub_video_fb_create_render_target_from_pointer (struct grub_video_fbrender_targ
 
   grub_memcpy (&(target->mode_info), mode_info, sizeof (target->mode_info));
 
-  /* Reset viewport to match new mode.  */
+  /* Reset viewport, region and area to match new mode.  */
   target->viewport.x = 0;
   target->viewport.y = 0;
   target->viewport.width = mode_info->width;
   target->viewport.height = mode_info->height;
 
+  target->region.x = 0;
+  target->region.y = 0;
+  target->region.width = mode_info->width;
+  target->region.height = mode_info->height;
+
+  target->area_enabled = 0;
+  target->area.x = 0;
+  target->area.y = 0;
+  target->area.width = mode_info->width;
+  target->area.height = mode_info->height;
+  target->area_offset_x = 0;
+  target->area_offset_y = 0;
+
   /* Clear render target with black and maximum transparency.  */
   for (y = 0; y < mode_info->height; y++)
     grub_memset (target->data + mode_info->pitch * y, 0,
